import numpy as np
from yggdrasil import units
from yggdrasil.communication.filters.FilterBase import FilterBase


class FunctionFilter(FilterBase):
    r"""Class for filtering messages based on a provided function using Python syntax.

    Args:
        function (func): The handle for a callable Python object (e.g. function)
            that should be used to determine if a message should be filtered or
            a string of the form "<function file>:<function name>" identifying
            a function where "<function file>" is the module or Python file
            containing the function and "<function name>" is the name of the function.
            The function should take the message as input and return a boolean, True
            if the message should pass through the filter, False if it should not.

    """
    _filtertype = 'function'
    _schema_required = ['function']
    _schema_properties = {'function': {'type': 'function'}}

    def evaluate_filter(self, x):
        r"""Call filter on the provided message.

        Args:
            x (object): Message object to filter.

        Returns:
            bool: True if the message will pass through the filter, False otherwise.

        """
        return self.function(x)
    
    @classmethod
    def get_testing_options(cls):
        r"""Get testing options for the filter class.

        Returns:
            list: Mutiple dictionaries of keywords and messages that will
                pass/fail for those keywords.
        
        """

        def fcond(x):
            return (units.get_data(x) != 3)
        
        return [{'kwargs': {'function': fcond},
                 'pass': [1, 2, units.add_units(1, 'cm'),
                          np.ones(3, int),
                          units.add_units(np.ones(3, int), 'cm')],
                 'fail': [3, units.add_units(3, 'cm'),
                          3 * np.ones(3, int),
                          units.add_units(3 * np.ones(3, int), 'cm')]}]



import pprint
from yggdrasil.examples.tests.test_transforms import TestExampleTransforms


def src():
    test_data = TestExampleTransforms.get_test_data()
    print('SENDING')
    pprint.pprint(test_data)
    return test_data



import copy
from yggdrasil.metaschema.datatypes import compare_schema
from yggdrasil.metaschema.datatypes.tests import (
    test_JSONObjectMetaschemaType as parent)


class TestSchemaMetaschemaType(parent.TestJSONObjectMetaschemaType):
    r"""Test class for SchemaMetaschemaType class."""

    _mod = 'SchemaMetaschemaType'
    _cls = 'SchemaMetaschemaType'

    @staticmethod
    def after_class_creation(cls):
        r"""Actions to be taken during class construction."""
        parent.TestJSONObjectMetaschemaType.after_class_creation(cls)
        cls._value = copy.deepcopy(cls._fulldef)
        cls._value['type'] = 'object'
        cls._fulldef = {'type': 'schema'}
        cls._typedef = {'type': 'schema'}
        cls._valid_encoded = [cls._fulldef]
        cls._valid_decoded = [cls._value]
        cls._invalid_validate = [None]
        cls._invalid_decoded = [{}]
        cls._invalid_encoded = [{}]
        cls._compatible_objects = [(cls._value, cls._value, None)]
        cls._valid_normalize += [('float', {'type': 'float',
                                            'precision': 64}),
                                 ({'units': 'g'}, {'units': 'g',
                                                   'type': 'scalar',
                                                   'subtype': 'float',
                                                   'precision': 64}),
                                 ({'title': 'a'}, {'title': 'a'}),
                                 ({'title': 'a', 'units': 'g'},
                                  {'title': 'a', 'units': 'g',
                                   'type': 'scalar', 'subtype': 'float',
                                   'precision': 64}),
                                 ({}, {})]

    @classmethod
    def assert_result_equal(cls, x, y):
        r"""Assert that serialized/deserialized objects equal."""
        compare_schema(x, y)



import copy
import numpy as np
from yggdrasil import serialize
from yggdrasil.tests import assert_equal
from yggdrasil.metaschema.datatypes.JSONObjectMetaschemaType import (
    JSONObjectMetaschemaType)
from yggdrasil.metaschema.datatypes.tests import test_MetaschemaType as parent
from yggdrasil.metaschema.datatypes.tests import (
    test_ContainerMetaschemaType as container_utils)


def test_coerce():
    r"""Test serialization of coerced types."""
    typedef = {'type': 'object',
               'properties': {'a': {'type': '1darray',
                                    'subtype': 'float',
                                    'title': 'a',
                                    'precision': 64}}}
    x = JSONObjectMetaschemaType(**typedef)
    key_order = ['a']
    msg_recv = {'a': np.zeros(3, 'float64')}
    msg_send_list = [{'a': np.zeros(3, 'float32')},
                     serialize.dict2numpy(msg_recv, order=key_order),
                     serialize.dict2pandas(msg_recv, order=key_order),
                     serialize.dict2list(msg_recv, order=key_order)]

    def do_send_recv(msg_send):
        msg_seri = x.serialize(msg_send, tyepdef=typedef, key_order=key_order)
        assert_equal(x.deserialize(msg_seri)[0], msg_recv)

    for y in msg_send_list:
        do_send_recv(y)


class TestJSONObjectMetaschemaType(parent.TestMetaschemaType):
    r"""Test class for JSONObjectMetaschemaType class."""

    _mod = 'JSONObjectMetaschemaType'
    _cls = 'JSONObjectMetaschemaType'

    @staticmethod
    def after_class_creation(cls):
        r"""Actions to be taken during class construction."""
        parent.TestMetaschemaType.after_class_creation(cls)
        cls._value = {}
        cls._fulldef = {'type': cls.get_import_cls().name,
                        'properties': {}}
        cls._typedef = {'properties': {}}
        for i, k in zip(range(container_utils._count), 'abcdefg'):
            cls._value[k] = container_utils._vallist[i]
            cls._fulldef['properties'][k] = container_utils._deflist[i]
            cls._typedef['properties'][k] = container_utils._typedef[i]
        cls._valid_encoded = [cls._fulldef]
        cls._valid_decoded = [cls._value]
        cls._invalid_encoded += [
            {'type': cls._fulldef['type'],
             'properties': {'a': cls._fulldef['properties']['a']}}]
        cls._invalid_encoded.append(copy.deepcopy(cls._fulldef))
        del cls._invalid_encoded[-1]['properties']['a']['type']
        cls._invalid_encoded.append(copy.deepcopy(cls._fulldef))
        cls._invalid_encoded[-1]['properties']['a']['type'] = 'invalid'
        cls._compatible_objects = [(cls._value, cls._value, None)]



import os
import copy
import shutil
import tempfile
import unittest
from yggdrasil.tests import assert_raises, assert_equal
from yggdrasil.metaschema.datatypes.tests import (
    test_PlyMetaschemaType as parent)
from yggdrasil.metaschema.datatypes import ObjMetaschemaType
from yggdrasil.drivers.LPyModelDriver import LPyModelDriver


old_value = parent._test_value
_test_value = {'vertices': [], 'faces': [], 'lines': [],
               'params': [{'u': 0.0, 'v': 0.0, 'w': 0.5},
                          {'u': 0.0, 'v': 0.0}],
               'normals': [{'i': 0.0, 'j': 0.0, 'k': 0.0},
                           {'i': 1.0, 'j': 1.0, 'k': 1.0}],
               'texcoords': [{'u': 0.0, 'v': 0.0, 'w': 0.5},
                             {'u': 1.0, 'v': 0.5},
                             {'u': 1.0}],
               'points': [[0, 2]],
               'curves': [{'starting_param': 0.0, 'ending_param': 1.0,
                           'vertex_indices': [0, 1]}],
               'curve2Ds': [[0, 1]],
               'surfaces': [{
                   'starting_param_u': 0.0, 'ending_param_u': 1.0,
                   'starting_param_v': 0.0, 'ending_param_v': 1.0,
                   'vertex_indices': [{'vertex_index': 0,
                                       'texcoord_index': 0,
                                       'normal_index': 0},
                                      {'vertex_index': 1,
                                       'texcoord_index': 1,
                                       'normal_index': 1}]}]}
_test_value['material'] = old_value['material']
_test_value['vertices'] = copy.deepcopy(old_value['vertices'])
_test_value['vertices'][0]['w'] = 0.5
for f in old_value['faces']:
    new = [{'vertex_index': x, 'texcoord_index': 0, 'normal_index': 0}
           for x in f['vertex_index']]
    _test_value['faces'].append(new)
for e in old_value['edges']:
    new = [{'vertex_index': e['vertex%d' % x]} for x in [1, 2]]
    _test_value['lines'].append(new)
_test_value_simple = {'vertices': copy.deepcopy(_test_value['vertices']),
                      'normals': copy.deepcopy(_test_value['normals']),
                      'texcoords': copy.deepcopy(_test_value['texcoords']),
                      'faces': [[{'vertex_index': 0, 'normal_index': 0},
                                 {'vertex_index': 1, 'normal_index': 0},
                                 {'vertex_index': 2, 'normal_index': 0}],
                                [{'vertex_index': 0, 'normal_index': 1},
                                 {'vertex_index': 2, 'normal_index': 1},
                                 {'vertex_index': 3, 'normal_index': 1}]]}
for f in _test_value_simple['faces']:
    for fv in f:
        fv['texcoord_index'] = 0
for v in _test_value_simple['vertices']:
    v.pop('w', None)
for t in _test_value_simple['texcoords']:
    t.pop('w', None)
    t.setdefault('v', 0.0)


def test_create_schema():
    r"""Test create_schema."""
    assert_raises(RuntimeError, ObjMetaschemaType.create_schema, overwrite=False)
    temp = os.path.join(tempfile.gettempdir(), 'temp_schema')
    old_schema = ObjMetaschemaType.get_schema()
    try:
        shutil.move(ObjMetaschemaType._schema_file, temp)
        new_schema = ObjMetaschemaType.get_schema()
        assert_equal(old_schema, new_schema)
    except BaseException:  # pragma: debug
        shutil.move(temp, ObjMetaschemaType._schema_file)
        raise
    shutil.move(temp, ObjMetaschemaType._schema_file)


class TestObjDict(parent.TestPlyDict):
    r"""Test for ObjDict class."""
    
    _mod = 'ObjMetaschemaType'
    _cls = 'ObjDict'
    _simple_test = _test_value_simple

    @property
    def inst_kwargs(self):
        r"""dict: Keyword arguments for creating a class instance."""
        return _test_value

    @classmethod
    def remove_hd_faces(cls, obj):
        r"""Remove higher dimension faces (>3 verts)."""
        if 'faces' in obj:
            obj['faces'] = [f[:3] for f in obj['faces']]
        return obj

    def test_apply_scalar_map(self):
        r"""Test applying a scalar colormap."""
        super(TestObjDict, self).test_apply_scalar_map(_as_obj=True)

    @unittest.skipIf(not LPyModelDriver.is_installed(), "LPy library not installed.")
    def test_to_from_scene(self, data=None):  # pragma: lpy
        r"""Test conversion to/from PlantGL scene."""
        super(TestObjDict, self).test_to_from_scene(_as_obj=True, data=data)

    def test_vertex_normals(self):
        r"""Test vertex_normals."""
        self.instance.vertex_normals

    def test_to_from_array_dict(self, test_objs=None):
        r"""Test transformation to/from dict of arrays."""
        if test_objs is None:
            orig = copy.deepcopy(self.instance)
            orig_edges = copy.deepcopy(self.instance)
            for v in orig_edges['texcoords']:
                v.pop('w', None)
            orig_nocolor = copy.deepcopy(self.instance)
            for v in orig_nocolor['vertices']:
                for k in ['red', 'green', 'blue']:
                    v.pop(k, None)
            for v in orig_nocolor['params']:
                v.pop('w', None)
            for v in orig_nocolor['texcoords']:
                v.pop('v', None)
                v.pop('w', None)
            for surf in orig_nocolor['surfaces']:
                for v in surf['vertex_indices']:
                    v.pop('texcoord_index', None)
                    v.pop('normal_index', None)
            test_objs = [orig, orig_edges, orig_nocolor]
        super(TestObjDict, self).test_to_from_array_dict(test_objs=test_objs)
            
    def test_to_from_trimesh(self, test_objs=None):
        r"""Test transformation to/from trimesh class."""
        if test_objs is None:
            orig = copy.deepcopy(self.instance)
            for k in ['material', 'params', 'points', 'surfaces', 'texcoords',
                      'curves', 'curve2Ds', 'lines', 'edges', 'normals']:
                orig.pop(k, None)
            for vlist in orig['faces']:
                for v in vlist:
                    v.pop('texcoord_index', None)
                    v.pop('normal_index', None)
            for v in orig['vertices']:
                for k in ['w']:  # , 'red', 'blue', 'green']:
                    v.pop(k, None)
            test_objs = [orig]
        super(TestObjDict, self).test_to_from_trimesh(test_objs=test_objs)
            
    @unittest.skipIf(not LPyModelDriver.is_installed(), "LPy library not installed.")
    def test_to_from_scene_incomplete(self):  # pragma: lpy
        r"""Test conversion to/from PlantGL scene with faces missing
        texcoords/normals on last element."""
        data = copy.deepcopy(self._simple_test)
        for f in data['faces']:
            for ff in f:
                ff.pop('texcoord_index', None)
                ff.pop('normal_index', None)
        self.test_to_from_scene(data=data)
        

class TestObjMetaschemaType(parent.TestPlyMetaschemaType):
    r"""Test class for ObjMetaschemaType class with float."""

    _mod = 'ObjMetaschemaType'
    _cls = 'ObjMetaschemaType'

    @staticmethod
    def after_class_creation(cls):
        r"""Actions to be taken during class construction."""
        parent.TestPlyMetaschemaType.after_class_creation(cls)
        cls._value = _test_value
        cls._fulldef = {'type': cls.get_import_cls().name}
        cls._typedef = {'type': cls._fulldef['type']}
        cls._valid_encoded = [cls._fulldef]
        cls._valid_decoded = [cls._value,
                              ObjMetaschemaType.ObjDict(**_test_value),
                              {'vertices': cls._value['vertices'],
                               'faces': [[{'vertex_index': 0},
                                          {'vertex_index': 1},
                                          {'vertex_index': 2}]]}]
        if ObjMetaschemaType.trimesh:
            cls._valid_decoded.append(
                ObjMetaschemaType.ObjDict(**_test_value).as_trimesh())
        # TODO: Add tests for faces with just texcoord or normal?
        cls._invalid_encoded = [{}]
        cls._invalid_decoded = [{'vertices': [{k: 0.0 for k in 'xyz'}],
                                 'faces': [[{'vertex_index': 0},
                                            {'vertex_index': 1},
                                            {'vertex_index': 2}]]},
                                {'vertices': [], 'faces': None,
                                 'lines': [[None]],
                                 'surfaces': [{'vertex_indices': [[]]}]},
                                {'vertices': cls._value['vertices'],
                                 'texcoords': cls._value['texcoords'],
                                 'normals': cls._value['normals'],
                                 'faces': [[{'vertex_index': 0,
                                             'texcoord_index': 100,
                                             'normal_index': 100},
                                            {'vertex_index': 1},
                                            {'vertex_index': 2}]]},
                                None]
        cls._compatible_objects = [(cls._value, cls._value, None)]

    @classmethod
    def assert_result_equal(cls, x, y):
        r"""Assert that serialized/deserialized objects equal."""
        if ObjMetaschemaType.trimesh:
            if isinstance(y, ObjMetaschemaType.trimesh.base.Trimesh):
                y = ObjMetaschemaType.ObjDict.from_trimesh(y)
        super(TestObjMetaschemaType, cls).assert_result_equal(x, y)



from yggdrasil.tests import generate_component_tests
from yggdrasil.serialize.tests.test_SerializeBase import TestSerializeBase


generate_component_tests('serializer', TestSerializeBase,
                         globals(), __file__)


__all__ = []



import os
import trimesh
import argparse
import numpy as np
from yggdrasil import units
from yggdrasil.languages.Python.YggInterface import YggRpcClient, YggOutput, YggTimesync

_dir = os.path.dirname(os.path.realpath(__file__))

def run(mesh, tmin, tmax, tstep):
    mass = 2000.0
    light_rpc = YggRpcClient('light_plant')
    light_out = YggOutput('light')
    plant2root = YggTimesync('plant2root')
    mass = units.add_units(mass, 'g')
    tmin = units.add_units(tmin, 'hrs')
    tmax = units.add_units(tmax, 'hrs')
    tstep = units.add_units(tstep, 'hrs')
    t = tmin
    i = 0

    while t <= tmax:

        # Perform send portion of call to synchronize data for time step
        # with the root model
        root_state = {'mass': mass}
        flag = plant2root.send(t, root_state)
        if not flag:
            raise Exception("Error performing time-step synchronization "
                            "with root model.")

        # Get light data by calling light model
        flag, light = light_rpc.call(mesh.vertices[:, 2], t)
        if not flag:
            raise Exception("Error calling light model")

        # Perform receive portion of call to synchronize data for time step
        # with the root model
        flag, root_state = plant2root.recv()
        if not flag:
            raise Exception("Error performing recv for time-step "
                            "synchronization with root model.")
        mass = root_state['mass']

        # Grow mesh
        # (pretend this is a biologically complex calculation)
        scale = units.get_data(mass * light / units.add_units(1.0, 'kg'))
        mesh.vertices[:, 2] += mesh.vertices[:, 2] * scale

        # Save mesh for this timestep
        filename_mesh = os.path.join(_dir, f'../output/mesh_{i:03d}.obj')
        with open(filename_mesh, 'w') as fd:
            mesh.export(fd, 'obj')

        # Send light to output
        flag = light_out.send(light)
        if not flag:
            raise Exception("Error sending light to output")

        # Advance time step
        t += tstep
        i += 1

    return mesh


# Parse command-line arguments
if __name__ == "__main__":
    parser = argparse.ArgumentParser("Simulate plant growth over time.")
    parser.add_argument('tmin', help='Starting time (in hours)', type=float)
    parser.add_argument('tmax', help='Ending time (in hours)', type=float)
    parser.add_argument('tstep', help='Time step (in hours)', type=float)
    parser.add_argument('--meshfile', help='Path to file where mesh is stored.',
                        default='../meshes/plants-2.obj')
    args = parser.parse_args()
    mesh = trimesh.load_mesh(args.meshfile)
    run(mesh, args.tmin, args.tmax, args.tstep)



# Import classes for input/output channels
from yggdrasil.interface.YggInterface import YggInput, YggOutput

# Initialize input/output channels
in_channel = YggInput('input')
out_channel = YggOutput('output')

# Loop until there is no longer input or the queues are closed
while True:
    
    # Receive input from input channel
    # If there is an error, the flag will be False
    flag, msg = in_channel.recv()
    if not flag:
        print("No more input.")
        break

    # Print received message
    print(msg)

    # Send output to output channel
    # If there is an error, the flag will be False
    flag = out_channel.send(msg)
    if not flag:
        print("Error sending output.")
        break



from yggdrasil import multitasking
from yggdrasil.communication import CommBase


class LockedBuffer(multitasking.Queue):
    r"""Buffer intended to be shared between threads/processes."""

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('task_method', 'process')
        super(LockedBuffer, self).__init__(*args, **kwargs)
        self._closed = self.context.Event()

    @property
    def closed(self):
        r"""bool: True if the queue is closed, False otherwise."""
        if hasattr(self, '_closed'):
            return self._closed.is_set()
        return True  # pragma: debug

    def close(self, join=False):
        r"""Close the buffer."""
        # with self.lock:
        self.disconnect()

    def disconnect(self):
        if hasattr(self, '_closed'):
            self._closed.set()
            self._closed.disconnect()
        super(LockedBuffer, self).disconnect()
        
    def __len__(self):
        if self.closed:  # pragma: debug
            return 0
        return int(not self.empty())

    def append(self, x):
        r"""Add an element to the queue."""
        self.put_nowait(x)

    def pop(self, index=None, default=None):
        r"""Remove the first element from the queue."""
        assert(index == 0)
        # with self.lock:
        if (len(self) == 0) and (default is not None):
            return default
        return self.get()

    def clear(self):
        r"""Remove all elements from the queue."""
        # with self.lock:
        while not self.empty():
            self.get()


class BufferComm(CommBase.CommBase):
    r"""Class for handling I/O to an in-memory buffer.

    Args:
        name (str): The name of the message queue.
        address (LockedBuffer, optional): Existing buffer that should be
            used. If not provided, a new buffer is created.
        buffer_task_method (str, optional): Type of tasks that buffer
            will be used to share information between. Defaults to 'thread'.
        **kwargs: Additional keyword arguments are passed to CommBase.
        
    Attributes:
        address (LockedBuffer): Buffer containing messages.
        buffer_task_method (str): Type of tasks that buffer will be
            used to share information between.
        
    """
    _commtype = 'buffer'
    no_serialization = True

    def __init__(self, *args, buffer_task_method="thread", **kwargs):
        self.buffer_task_method = buffer_task_method
        super(BufferComm, self).__init__(*args, **kwargs)

    @classmethod
    def is_installed(cls, language=None):
        r"""Determine if the necessary libraries are installed for this
        communication class.

        Args:
            language (str, optional): Specific language that should be checked
                for compatibility. Defaults to None and all languages supported
                on the current platform will be checked. If set to 'any', the
                result will be True if this comm is installed for any of the
                supported languages.

        Returns:
            bool: Is the comm installed.

        """
        if language == 'python':
            return True
        return False
        
    def bind(self):
        r"""Bind to address, getting random port as necessary."""
        super(BufferComm, self).bind()
        if not isinstance(self.address, LockedBuffer):
            if self.address == 'address':
                self.address = LockedBuffer(task_method=self.buffer_task_method)
            else:  # pragma: debug
                raise ValueError("Invalid address for a buffer: %s"
                                 % self.address)
        
    def serialize(self, args, **kwargs):
        r"""Serialize a message using the associated serializer."""
        return args
        
    def deserialize(self, args, **kwargs):
        r"""Deserialize a message using the associated deserializer."""
        return args, {}
    
    @property
    def is_open(self):
        r"""bool: True if the connection is open."""
        return (not self.address.closed)
        
    def _close(self, *args, **kwargs):
        r"""Close the connection."""
        self.address.close()
        
    @property
    def n_msg_recv(self):
        r"""int: Number of messages in the receive backlog."""
        return len(self.address)

    @property
    def n_msg_send(self):
        r"""int: Number of messages in the send backlog."""
        return len(self.address)

    @property
    def is_confirmed_send(self):
        r"""bool: True if all sent messages have been confirmed."""
        return (self.n_msg_send == 0)

    @property
    def is_confirmed_recv(self):
        r"""bool: True if all received messages have been confirmed."""
        return (self.n_msg_recv == 0)

    def _send(self, payload, **kwargs):
        r"""Send a message to the buffer.

        Args:
            payload (str): Message to send.

        Returns:
            bool: Success or failure of sending the message.

        """
        if self.direction == 'recv':  # pragma: debug
            self.error("Receiving buffer comm cannot send.")
            return False
        self.address.append(payload)
        return True

    def _recv(self, timeout=None):
        r"""Receive a message from the buffer.

        Args:
            timeout (float, optional): Time in seconds to wait for a message.
                Defaults to self.recv_timeout.

        Returns:
            tuple (bool, str): The success or failure of receiving a message
                and the message received.

        """
        if timeout is None:  # pragma: no cover
            timeout = self.recv_timeout
        if self.direction == 'send':  # pragma: debug
            self.error("Sending buffer comm cannot receive.")
            return (False, self.empty_bytes_msg)
        # Sleep until there is a message
        T = self.start_timeout(timeout, key_suffix='_recv')
        while (not T.is_out) and (not len(self.address)):
            self.sleep()
        self.stop_timeout(key_suffix='_recv')
        return (True, self.address.pop(0, self.empty_bytes_msg))

    def purge(self):
        r"""Purge all messages from the comm."""
        super(BufferComm, self).purge()
        self.address.clear()



import os
from yggdrasil.examples.tests import ExampleTstBase


class TestTimedPipeBase(ExampleTstBase):
    r"""Base class for testing TimedPipe example with various comm types."""

    example_name = 'timed_pipe'
    env = {'PIPE_MSG_COUNT': '10',
           'PIPE_MSG_SIZE': '1024'}
    iter_over = ['language', 'comm']

    @property
    def description_prefix(self):
        r"""Prefix message with test name."""
        out = super(TestTimedPipeBase, self).description_prefix
        out += '(%s)' % self.comm
        return out
    
    @property
    def results(self):
        r"""Result that should be found in output files."""
        siz = int(self.env['PIPE_MSG_COUNT']) * int(self.env['PIPE_MSG_SIZE'])
        res = '0' * siz
        return [res]

    @property
    def output_files(self):
        r"""Output file."""
        return [os.path.join(self.tempdir, 'output_timed_pipe.txt')]



import os
import pprint
from yggdrasil.examples.tests.test_types import TestExampleTypes


def src():
    typename = os.environ.get('TEST_TYPENAME', None)
    assert(typename is not None)
    test_data = TestExampleTypes.get_test_data(typename)
    print('SENDING: %s' % typename)
    pprint.pprint(test_data)
    return test_data



import pandas
import copy
import numpy as np
import warnings
import io as sio
from yggdrasil import platform, serialize, units
from yggdrasil.metaschema.datatypes.JSONArrayMetaschemaType import (
    JSONArrayMetaschemaType)
from yggdrasil.serialize.AsciiTableSerialize import AsciiTableSerialize
from yggdrasil.communication.transforms.PandasTransform import PandasTransform


class PandasSerialize(AsciiTableSerialize):
    r"""Class for serializing/deserializing Pandas data frames.

    Args:
        no_header (bool, optional): If True, headers will not be read or
            serialized from/to tables. Defaults to False.
        str_as_bytes (bool, optional): If True, strings in columns are
            read as bytes. Defaults to False.

    """

    _seritype = 'pandas'
    _schema_subtype_description = ('Serializes tables using the pandas package.')
    _schema_properties = {'no_header': {'type': 'boolean',
                                        'default': False},
                          'str_as_bytes': {'type': 'boolean',
                                           'default': False}}
    _schema_excluded_from_inherit = ['as_array']
    default_read_meth = 'read'
    as_array = True
    concats_as_str = False
    # has_header = False

    def __init__(self, *args, **kwargs):
        self.write_header_once = False
        self.dont_write_header = kwargs.pop('dont_write_header',
                                            kwargs.get('no_header', False))
        return super(PandasSerialize, self).__init__(*args, **kwargs)

    @property
    def empty_msg(self):
        r"""obj: Object indicating empty message."""
        if self.numpy_dtype:
            return pandas.DataFrame(np.zeros(0, self.numpy_dtype))
        else:
            return pandas.DataFrame(columns=self.get_field_names())

    def get_field_names(self, *args, **kwargs):
        r"""Get the field names for an array of fields.

        Args:
            *args: Arguments are passed to the parent class's method.
            **kwargs: Keyword arguments are passed to the parent class's
                method.

        Returns:
            list: Names for each field in the data type.

        """
        if self.no_header:
            return None
        return super(PandasSerialize, self).get_field_names(*args, **kwargs)
        
    @classmethod
    def apply_field_names(cls, frame, field_names=None):
        r"""Apply field names as columns to a frame, first checking for a mapping.
        If there is a direct mapping, the columns are reordered to match the order
        of the field names. If there is not an overlap in the field names and
        columns, a one-to-one mapping is assumed, but a warning is issued. If there
        is a partial overlap, an error is raised.

        Args:
            frame (pandas.DataFrame): Frame to apply field names to as columns.
            field_names (list, optional): New field names that should be applied.
                If not provided, the original frame will be returned unaltered.

        Returns:
            pandas.DataFrame: Frame with updated field names.

        Raises:
            RuntimeError: If there is a partial overlap between the field names
                and columns.

        """
        if field_names is None:
            return frame
        cols = frame.columns.tolist()
        if len(field_names) != len(cols):
            raise RuntimeError(("Number of field names (%d) doesn't match "
                                + "number of columns in data frame (%d).")
                               % (len(field_names), len(cols)))
        # Check for missing fields
        fmiss = []
        for f in field_names:
            if f not in cols:
                fmiss.append(f)
        if fmiss:
            if len(fmiss) == len(field_names):
                warnings.warn("Assuming direct mapping of field names to columns. "
                              + "This may not be correct.")
                frame.columns = field_names
            else:
                # Partial overlap
                raise RuntimeError("%d fields (%s) missing from frame: %s"
                                   % (len(fmiss), str(fmiss), str(frame)))
        else:
            # Reorder columns
            frame = frame[field_names]
        return frame

    def cformat2nptype(self, *args, **kwargs):
        r"""Method to convert c format string to numpy data type.

        Args:
            *args: Arguments are passed to serialize.cformat2nptype.
            **kwargs: Keyword arguments are passed to serialize.cformat2nptype.

        Returns:
            np.dtype: Corresponding numpy data type.

        """
        out = super(PandasSerialize, self).cformat2nptype(*args, **kwargs)
        if (out.char == 'S') and (not self.str_as_bytes):
            out = np.dtype('U%d' % out.itemsize)
        return out
    
    def func_serialize(self, args):
        r"""Serialize a message.

        Args:
            args (obj): Python object to be serialized.

        Returns:
            bytes, str: Serialized message.

        """
        if not isinstance(args, pandas.DataFrame):
            raise TypeError(("Pandas DataFrame required. Invalid type "
                             + "of '%s' provided.") % type(args))
        fd = sio.StringIO()
        # For Python 3 and higher, bytes need to be encoded
        args_ = copy.deepcopy(args)
        for c in args.columns:
            if isinstance(args_[c][0], bytes):
                args_[c] = args_[c].apply(lambda s: s.decode('utf-8'))
        if (self.field_names is None) and (not self.no_header):
            self.field_names = self.get_field_names()
        args_ = self.apply_field_names(args_, self.field_names)
        if not self.no_header:
            cols = args_.columns.tolist()
            if cols == list(range(len(cols))):
                args_ = self.apply_field_names(args_, ['f%d' % i for i in
                                                       range(len(cols))])
        args_.to_csv(fd, index=False,
                     # Not in pandas <0.24
                     # line_terminator=self.newline.decode("utf-8"),
                     sep=self.delimiter.decode("utf-8"),
                     mode='wb', encoding='utf8',
                     header=(not self.dont_write_header))
        if self.write_header_once:
            self.dont_write_header = True
        out = fd.getvalue()
        fd.close()
        # Required to change out \r\n for \n on windows
        out = out.encode("utf-8")
        out = out.replace(platform._newline, self.newline)
        return out

    def func_deserialize(self, msg):
        r"""Deserialize a message.

        Args:
            msg (str, bytes): Message to be deserialized.

        Returns:
            obj: Deserialized Python object.

        """
        fd = sio.BytesIO(msg)
        names = None
        dtype = None
        if self.initialized:
            np_dtype = self.numpy_dtype
            dtype = {}
            if self.no_header:
                dtype_names = range(len(np_dtype.names))
            else:
                dtype_names = np_dtype.names
            for n in dtype_names:
                if np_dtype[n].char == 'U':
                    dtype[n] = object
                else:
                    dtype[n] = np_dtype[n]
        kws = dict(sep=self.delimiter.decode("utf-8"),
                   names=names,
                   dtype=dtype,
                   encoding='utf8',
                   skipinitialspace=True)
        if self.no_header:
            kws['header'] = None
        out = pandas.read_csv(fd, **kws)
        out = out.dropna(axis='columns', how='all')
        fd.close()
        if self.str_as_bytes:
            # Make sure strings are bytes
            for c, d in zip(out.columns, out.dtypes):
                if (d == object) and isinstance(out[c][0], str):
                    out[c] = out[c].apply(lambda s: s.encode('utf-8'))
        # On windows, long != longlong and longlong requires special cformat
        # For now, long will be used to preserve the use of %ld to match long
        if platform._is_win:  # pragma: windows
            if np.dtype('longlong').itemsize == 8:
                new_dtypes = dict()
                for c, d in zip(out.columns, out.dtypes):
                    if d == np.dtype('longlong'):
                        new_dtypes[c] = np.int32
                    else:
                        new_dtypes[c] = d
                out = out.astype(new_dtypes, copy=False)
        # Reorder if necessary
        out = self.apply_field_names(out, self.get_field_names())
        if dtype is not None:
            out = out.astype(dtype, copy=False)
        if (self.field_names is None) and (not self.no_header):
            self.field_names = out.columns.tolist()
        if not self.initialized:
            typedef = JSONArrayMetaschemaType.encode_type(out)
            self.update_serializer(extract=True, **typedef)
        return out

    @property
    def send_converter(self):
        kws = {}
        field_names = self.get_field_names()
        if field_names is not None:
            kws['field_names'] = field_names
        return PandasTransform(**kws)

    @classmethod
    def object2dict(cls, obj, **kwargs):
        r"""Convert a message object into a dictionary.

        Args:
            obj (object): Object that would be serialized by this class and
                should be returned in a dictionary form.
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            dict: Dictionary version of the provided object.

        """
        if isinstance(obj, pandas.DataFrame):
            return serialize.pandas2dict(obj)
        return super(PandasSerialize, cls).object2dict(obj, as_array=True,
                                                       **kwargs)

    @classmethod
    def object2array(cls, obj, **kwargs):
        r"""Convert a message object into an array.

        Args:
            obj (object): Object that would be serialized by this class and
                should be returned in an array form.
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            np.array: Array version of the provided object.

        """
        if isinstance(obj, pandas.DataFrame):
            return serialize.pandas2numpy(obj)
        return super(PandasSerialize, cls).object2array(obj, as_array=True,
                                                        **kwargs)

    @classmethod
    def concatenate(cls, objects, **kwargs):
        r"""Concatenate objects to get object that would be recieved if
        the concatenated serialization were deserialized.

        Args:
            objects (list): Objects to be concatenated.
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            list: Set of objects that results from concatenating those provided.

        """
        if len(objects) == 0:
            return []
        if isinstance(objects[0], pandas.DataFrame):
            field_names = objects[0].columns.tolist()
            for i in range(1, len(objects)):
                objects[i] = cls.apply_field_names(objects[i],
                                                   field_names)
            return [pandas.concat(objects, ignore_index=True)]
        out = super(PandasSerialize, cls).concatenate(objects, as_array=True,
                                                      **kwargs)
        return out
    
    def consolidate_array(self, out):
        r"""Consolidate message into a structure numpy array if possible.

        Args:
            out (list, tuple, np.ndarray): Object to consolidate into a
                structured numpy array.

        Returns:
            np.ndarray: Structured numpy array containing consolidated message.

        Raises:
            ValueError: If the array cannot be consolidated.

        """
        if isinstance(out, pandas.DataFrame):
            out = serialize.pandas2numpy(out)
        return super(PandasSerialize, self).consolidate_array(out)
        
    @classmethod
    def get_testing_options(cls, not_as_frames=False, no_names=False,
                            no_header=False, **kwargs):
        r"""Method to return a dictionary of testing options for this class.

        Args:
            not_as_frames (bool, optional): If True, the returned example
                includes data that is not in a pandas data frame. Defaults to
                False.
            no_names (bool, optional): If True, an example is returned where the
                names are not provided to the deserializer. Defaults to False.
            no_header (bool, optional): If True, an example is returned
            where a header is not included. Defaults to False.

        Returns:
            dict: Dictionary of variables to use for testing.

        """
        kwargs.setdefault('table_string_type', 'string')
        field_names = None
        out = super(PandasSerialize, cls).get_testing_options(array_columns=True,
                                                              **kwargs)
        if kwargs['table_string_type'] == 'bytes':
            out['kwargs']['str_as_bytes'] = True
        for k in ['as_array']:  # , 'format_str']:
            if k in out['kwargs']:
                del out['kwargs'][k]
        out['extra_kwargs'] = {}
        if no_names:
            for x in [out['kwargs'], out]:
                if 'field_names' in x:
                    del x['field_names']
            header_line = b'f0\tf1\tf2\n'
        elif no_header:
            for x in [out['kwargs'], out]:
                if 'field_names' in x:
                    del x['field_names']
            header_line = b''
            out['kwargs']['no_header'] = True
            for x in out['typedef']['items']:
                x.pop('title', None)
        else:
            if 'field_names' in out['kwargs']:
                field_names = out['kwargs']['field_names']
            header_line = b'name\tcount\tsize\n'
        out['contents'] = (header_line
                           + b'one\t1\t1.0\n'
                           + b'two\t2\t2.0\n'
                           + b'three\t3\t3.0\n'
                           + b'one\t1\t1.0\n'
                           + b'two\t2\t2.0\n'
                           + b'three\t3\t3.0\n')
        out['concatenate'] = [([], [])]
        if not_as_frames:
            # Strip units since pandas data frames are not serialized with units
            out['objects'] = [[units.get_data(ix) for ix in x]
                              for x in out['objects']]
        elif no_header:
            out['objects'] = [serialize.list2pandas(x) for x in out['objects']]
            out['dtype'] = np.dtype(','.join([x[1] for x in out['dtype'].descr]))
        else:
            if field_names is None:
                field_names = ['f0', 'f1', 'f2']
            out['objects'] = [serialize.list2pandas(x, names=field_names)
                              for x in out['objects']]
        out['kwargs']['datatype'] = copy.deepcopy(out['typedef'])
        if no_names:
            for x in out['kwargs']['datatype']['items']:
                x.pop('title', None)
        out['empty'] = pandas.DataFrame(np.zeros(0, out['dtype']))
        return out

    def enable_file_header(self):
        r"""Set serializer attributes to enable file headers to be included in
        the serializations."""
        self.dont_write_header = False
        self.write_header_once = True

    def disable_file_header(self):
        r"""Set serializer attributes to disable file headers from being
        included in the serializations."""
        self.dont_write_header = True
        self.write_header_once = True
        
    def serialize_file_header(self):
        r"""Return the serialized header information that should be prepended
        to files serialized using this class.

        Returns:
            bytes: Header string that should be written to the file.

        """
        return b''

    def deserialize_file_header(self, fd):
        r"""Deserialize the header information from the file and update the
        serializer.

        Args:
            fd (file): File containing header.

        """
        pass



import numpy as np
from yggdrasil import units


def light(height, time):
    intensity = (
        80.0 *
        (1.0 + np.sin(2.0 * np.pi * time / units.add_units(365.0, 'days'))) /
        (np.abs(200 - height)**2))
    return intensity



import pickle
from yggdrasil import multitasking, tools
from yggdrasil.tests import assert_raises
from yggdrasil.tests.test_tools import YggTestClass


class TstClass(object):
    
    x = 'test'


class LockedTstClass(multitasking.LockedObject):

    _base_class = TstClass
    _base_attr = ['x']


def target():
    raise RuntimeError("Test error.")


def test_LockedAttr():
    r"""Test access to locked attribute."""
    z = LockedTstClass()
    z.x
    z.disconnect()
    assert_raises(multitasking.AliasDisconnectError,
                  getattr, z, 'x')


def test_TaskThread():
    r"""Test thread based Task."""
    q = multitasking.Task(target=target)
    assert(not q.is_alive())
    q.start()
    q._base._errored.wait(1.0)
    assert(q._base._errored.is_set())


def test_TaskProcess():
    r"""Test process based Task."""
    q = multitasking.Task(target=target, task_method='parallel')
    assert(not q.is_alive())
    q.start()
    q.join(60.0)
    assert(not q.is_alive())


class TestContextThread(YggTestClass):
    r"""Test for thread based Context."""

    _cls = 'Context'
    _mod = 'yggdrasil.multitasking'
    _task_method = 'thread'

    def __init__(self, *args, **kwargs):
        super(TestContextThread, self).__init__(*args, **kwargs)
        self._inst_kwargs = {'task_method': self._task_method}

    def test_RLock(self):
        r"""Test creation of RLock from context."""
        x = self.instance.RLock()
        x.disconnect()

    def test_Event(self):
        r"""Test creation of Event from context."""
        x = self.instance.Event()
        x.disconnect()

    def test_Task(self):
        r"""Test creation of Task from context."""
        x = self.instance.Task()
        x.disconnect()

    def test_Queue(self):
        r"""Test creation of Queue from context."""
        x = self.instance.Queue()
        x.disconnect()


class TestContextProcess(YggTestClass):
    r"""Test for process based Context."""

    _task_method = 'process'


class TstContextObject(object):

    _cls = None
    _mod = 'yggdrasil.multitasking'

    def check_decoded(self, decoded):
        r"""Check that object was decoded correctly."""
        pass

    def test_pickle(self):
        r"""Test pickling and unpickling the object."""
        encoded = pickle.dumps(self.instance)
        decoded = pickle.loads(encoded)
        self.check_decoded(decoded)


class TestRLock(TstContextObject, YggTestClass):

    _cls = 'RLock'


class TestEvent(TstContextObject, YggTestClass):

    _cls = 'Event'

    def check_decoded(self, decoded):
        r"""Check that object was decoded correctly."""
        self.assert_equal(decoded.is_set(), self.instance.is_set())

    def test_pickle(self):
        r"""Test pickling and unpickling the object."""
        self.instance.set()
        super(TestEvent, self).test_pickle()


class TestTask(TstContextObject, YggTestClass):

    _cls = 'Task'

    def check_decoded(self, decoded):
        r"""Check that object was decoded correctly."""
        self.assert_equal(decoded.name, self.instance.name)
        self.assert_equal(decoded.daemon, self.instance.daemon)


class TestQueue(TstContextObject, YggTestClass):

    _cls = 'Queue'

    def test_join(self):
        r"""Test join."""
        self.instance.join()


class TestYggTask(YggTestClass):
    r"""Test basic behavior of YggTask class."""

    _cls = 'YggTask'
    _mod = 'yggdrasil.multitasking'

    def __init__(self, *args, **kwargs):
        super(TestYggTask, self).__init__(*args, **kwargs)
        self.namespace = 'TESTING_%s' % self.uuid
        self.attr_list += ['name', 'sleeptime', 'longsleep', 'timeout']
        self._inst_kwargs = {'timeout': self.timeout,
                             'sleeptime': self.sleeptime,
                             'target': self.target}
        self.debug_flag = False

    @staticmethod
    def target(cls):  # pragma: no cover
        tools.sleep(10.0)

    def test_id(self):
        r"""Test process ID and ident."""
        self.instance.pid
        self.instance.ident

    def test_daemon(self):
        r"""Test process/thread daemon property."""
        self.instance.daemon

    def test_exitcode(self):
        r"""Test process exitcode."""
        self.instance.exitcode

    def test_get_main_proc(self):
        r"""Test get_main_proc."""
        self.instance.get_main_proc()

    def test_kill(self):
        r"""Test kill."""
        if self._inst_kwargs.get('task_method', None) == 'process':
            self.instance.start()
        self.instance.kill()

    def test_flag_manipulation(self):
        r"""Test flag manipulation."""
        self.instance.set_flag_attr('error_flag')
        assert(self.instance.check_flag_attr('error_flag'))
        self.instance.clear_flag_attr('error_flag')
        assert(not self.instance.check_flag_attr('error_flag'))


class TestYggProcess(TestYggTask):
    r"""Test basic behavior of YggTask for spawned process."""

    def __init__(self, *args, **kwargs):
        super(TestYggProcess, self).__init__(*args, **kwargs)
        self._inst_kwargs['task_method'] = 'process'


class TestYggProcessFork(TestYggProcess):
    r"""Test basic behavior of YggTask for forked process."""

    def __init__(self, *args, **kwargs):
        super(TestYggProcessFork, self).__init__(*args, **kwargs)
        self._inst_kwargs['context'] = multitasking.mp_ctx



import os
import six
import unittest
from yggdrasil import yamlfile, runner, tools
from yggdrasil.components import ComponentMeta, import_component
from yggdrasil.tests import YggTestBase, check_enabled_languages


_demo_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


class DemoRun(object):

    def __init__(self, name, yamls):
        self.runner = None
        self.name = name
        self.yamls = yamls
        self.languages = []
        for x in yamlfile.parse_yaml(yamls)['model'].values():
            drv = import_component('model', x['driver'],
                                   without_schema=True)
            if drv.language not in self.languages:
                self.languages.append(drv.language)
        super(DemoRun, self).__init__()

    def run(self):
        r"""Run the integration."""
        if not tools.check_environ_bool('YGG_ENABLE_DEMO_TESTS'):
            raise unittest.SkipTest("Demo tests not enabled.")
        for x in self.languages:
            check_enabled_languages(x)
            if not tools.is_lang_installed(x):
                raise unittest.SkipTest("%s not installed." % x)
        self.runner = runner.get_runner(self.yamls, namespace=self.name,
                                        production_run=True)
        self.runner.run()
        assert(not self.runner.error_flag)
        self.runner = None

    def add_test(self, dct):
        def itest(solf):
            self.run()
        itest.__name__ = 'test_%s' % self.name
        dct[itest.__name__] = itest


class DemoMeta(ComponentMeta):

    def __new__(cls, name, bases, dct):
        runs = dct.get('runs', {})
        for name, yamls in runs.items():
            runs[name] = DemoRun(name, [os.path.join(_demo_dir,
                                                     dct['demo_name'], iyml)
                                        for iyml in yamls])
            runs[name].add_test(dct)
        dct['runs'] = runs
        out = super(DemoMeta, cls).__new__(cls, name, bases, dct)
        return out


@six.add_metaclass(DemoMeta)
class DemoTstBase(YggTestBase, tools.YggClass):
    r"""Base class for running demos."""

    demo_name = None
    runs = {}



"""Tools for accessing examples from python."""
import os
import glob
import logging
from yggdrasil import tools, languages, serialize, constants


# TODO: This can be generated from the drivers
ext_map = dict(constants.LANG2EXT, executable='',
               make='.cpp', cmake='.cpp',
               osr='.xml', sbml='.xml')
for lang in tools.get_supported_lang():
    if lang.lower() not in ext_map:
        ext_map[lang.lower()] = languages.get_language_ext(lang)
_example_dir = os.path.dirname(__file__)


def register_example(example_dir):
    r"""Register an example based on the contents of the director.

    Args:
        example_dir (str): Full path to a directory potentially containing an
            example.

    Returns:
        tuple (list, dict, dict): A list of available languages, a dictionary
            mapping from language to YAML specification files for the example,
            and a dictionary mapping from language to source files for the
            example.

    """
    # Check that the source directory and test exist
    example_base = os.path.basename(example_dir)
    srcdir = os.path.join(example_dir, 'src')
    testfile = os.path.join(_example_dir, 'tests',
                            'test_%s.py' % example_base)
    if not os.path.isfile(testfile):  # pragma: no cover
        # TODO: Automate test creation
        if not tools.is_subprocess():
            logging.error("Missing test file: %s" % testfile)
    if not os.path.isdir(srcdir):  # pragma: no cover
        if not tools.is_subprocess():
            logging.error("Missing source directory: %s" % srcdir)
        return {}
    # Determine which languages are present in the example
    lang_base = []
    lang_avail = []
    lang_search = None
    if example_base in ['rpcFib', 'maxMsg']:
        lang_search = example_base + 'Cli_%s.yml'
        lang_avail += ['all', 'all_nomatlab']
    elif example_base.startswith('rpc_'):
        lang_search = 'client_%s.yml'
    elif example_base == 'root_to_shoot':
        lang_avail += ['all', 'all_nomatlab', 'c', 'python']
    elif example_base == 'fakeplant':
        lang_avail += ['all', 'all_nomatlab', 'c', 'cpp', 'matlab', 'python']
    elif example_base in ['types', 'transforms']:
        lang_avail += tools.get_supported_lang()
        for k in ['cmake', 'make', 'lpy', 'executable']:
            lang_avail.remove(k)
    elif example_base.startswith('sbml'):
        lang_avail = ['sbml']
    elif example_base.startswith('osr'):
        lang_search = example_base + '_%s.yml'
        lang_base += ['osr']
    else:
        lang_search = example_base + '_%s.yml'
    if lang_search is not None:
        for match in glob.glob(os.path.join(example_dir, lang_search % '*')):
            lang = serialize.process_message(os.path.basename(match), lang_search)[0]
            if lang == 'valgrind':
                continue
            lang_avail.append(lang.lower())
    lang_avail = tuple(sorted(lang_avail))
    lang_base = tuple(sorted(lang_base))
    # Get YAML and source files for each language
    out_yml = {}
    out_src = {}
    src_is_abs = False
    for lang in lang_avail:
        yml_names = []
        src_names = []
        if example_base == 'rpcFib':
            if lang == 'all':
                lang_set = ('python', 'matlab', 'c')
            elif lang == 'all_nomatlab':
                lang_set = ('python', 'cpp', 'c')
            else:
                lang_set = (lang, lang, lang)
            yml_names = ['%sCli_%s.yml' % (example_base, lang_set[0]),
                         '%sCliPar_%s.yml' % (example_base, lang_set[1]),
                         '%sSrv_%s.yml' % (example_base, lang_set[2])]
            src_names = ['%sCli%s' % (example_base, ext_map[lang_set[0]]),
                         '%sCliPar%s' % (example_base, ext_map[lang_set[1]]),
                         '%sSrv%s' % (example_base, ext_map[lang_set[2]])]
        elif example_base == 'maxMsg':
            if lang == 'all':
                lang_set = ('python', 'matlab')
            elif lang == 'all_nomatlab':
                lang_set = ('python', 'c')
            else:
                lang_set = (lang, lang)
            yml_names = ['%sCli_%s.yml' % (example_base, lang_set[0]),
                         '%sSrv_%s.yml' % (example_base, lang_set[1])]
            src_names = ['%sCli%s' % (example_base, ext_map[lang_set[0]]),
                         '%sSrv%s' % (example_base, ext_map[lang_set[1]])]
        elif example_base.startswith('rpc_'):
            # TODO: Create server examples in other languages
            yml_names = ['server_python.yml',
                         'client_%s.yml' % lang]
            src_names = ['server.py', 'client%s' % ext_map[lang]]
        elif example_base == 'root_to_shoot':
            if lang.startswith('all'):
                yml_names = ['root.yml', 'shoot.yml', 'root_to_shoot.yml']
                src_names = ['root.c', 'shoot.py']
            elif lang == 'python':
                yml_names = ['shoot.yml', 'shoot_files.yml']
                src_names = ['shoot.py']
            elif lang == 'c':
                yml_names = ['root.yml', 'root_files.yml']
                src_names = ['root.c']
        elif example_base == 'fakeplant':
            if lang.startswith('all'):
                yml_names = ['canopy.yml', 'light.yml', 'photosynthesis.yml',
                             'fakeplant.yml']
                src_names = ['canopy.cpp', 'light.c', 'photosynthesis.py']
                if lang == 'all_nomatlab':
                    yml_names.append('growth_python.yml')
                    src_names.append('growth.py')
                else:
                    yml_names.append('growth.yml')
                    src_names.append('growth.m')
            elif lang == 'python':
                yml_names = ['photosynthesis.yml']
                src_names = ['photosynthesis.py']
            elif lang == 'c':
                yml_names = ['light.yml', 'light_files.yml']
                src_names = ['light.c']
            elif lang == 'cpp':
                yml_names = ['canopy.yml', 'canopy_files.yml']
                src_names = ['canopy.cpp']
            elif lang == 'matlab':
                yml_names = ['growth.yml', 'growth_files.yml']
                src_names = ['growth.m']
        elif example_base in ['types', 'transforms']:
            yml_names = ['%s.yml' % example_base]
            src_names = ['src.py', 'dst.py']
        elif example_base.startswith('sbml'):
            yml_names = ['%s.yml' % example_base]
            src_names = ['%s.xml' % example_base]
        else:
            src_is_abs = True
            yml_names = ['%s_%s.yml' % (example_base, lang)]
            if lang.startswith('all'):
                src_names = []
                for lsrc in lang_avail:
                    if lsrc.startswith('all'):
                        continue
                    elif (lang == 'all_nomatlab') and (lsrc == 'matlab'):
                        continue
                    src_names += sorted(
                        glob.glob(os.path.join(srcdir,
                                               '*' + ext_map[lsrc])))
            else:
                src_names = sorted(
                    glob.glob(os.path.join(
                        srcdir, '*' + ext_map.get(lang, ''))))
            for ilang_base in lang_base:
                src_names += sorted(
                    glob.glob(os.path.join(
                        srcdir, '*' + ext_map.get(ilang_base, ''))))
        out_yml[lang] = [os.path.join(example_dir, y) for y in yml_names]
        if src_is_abs:
            out_src[lang] = src_names
        else:
            out_src[lang] = [os.path.join(srcdir, s) for s in src_names]
        if len(out_yml[lang]) == 1:
            out_yml[lang] = out_yml[lang][0]
        if len(out_src[lang]) == 1:
            out_src[lang] = out_src[lang][0]
    return lang_base, lang_avail, out_yml, out_src


def discover_examples(parent_dir=None):
    r"""Discover examples under the provided parent directory.

    Args:
        parent_dir (str, optional): Parent directory containing example
            directories. Defaults to the directory containing this function if
            not provided.

    Returns:

    Raises:

    """
    out = {'base_lang': {}, 'lang': {}, 'yml': {}, 'src': {}}
    if parent_dir is None:
        parent_dir = _example_dir
    for match in sorted(glob.glob(os.path.join(parent_dir, '*'))):
        if match.endswith(('tests', '_')) or (not os.path.isdir(match)):
            continue
        match_base = os.path.basename(match)
        iout = register_example(match)
        if iout:
            for i, k in enumerate(['base_lang', 'lang', 'yml', 'src']):
                out[k][match_base] = iout[i]
    return out['base_lang'], out['lang'], out['yml'], out['src']


base_langs, avail_langs, yamls, source = discover_examples()


def get_example_yaml(name, language):
    r"""Get yaml file(s) associated with an example in a certain language.

    Args:
        name (str): Name of the example.
        language (str): Language for example version that should be returned.

    Returns:
        str, list: One or more yaml file(s) associated with the example.

    """
    if name not in yamls:
        raise KeyError("Could not locate yaml for example: '%s'" % name)
    if language in yamls[name]:
        return yamls[name][language]
    elif language.lower() in yamls[name]:
        return yamls[name][language.lower()]
    raise KeyError("Could not locate yaml for example '%s' in language '%s'"
                   % (name, language))


def get_example_source(name, language):
    r"""Get source file(s) associated with an example in a certain language.

    Args:
        name (str): Name of the example.
        language (str): Language for example version that should be returned.

    Returns:
        str, list: One or more source file(s) associated with the example.

    """
    if name not in source:
        raise KeyError("Could not locate source for example: '%s'" % name)
    if language in source[name]:
        return source[name][language]
    elif language.lower() in source[name]:
        return source[name][language.lower()]
    raise KeyError("Could not locate source for example '%s' in language '%s'"
                   % (name, language))


def get_example_languages(name, language=None):
    r"""Get the set of languages that the example is available in.

    Args:
        name (str): Name of the example.
        language (str, optional): If provided, the returned languages will be
            those tested by the version of the example in the specified
            language. Defaults to None and the languages the example is
            available in will be returned.

    Returns:
        list: Languages that the example is available in.

    """
    if name not in avail_langs:
        raise KeyError("Could not locate languages for example: '%s'" % name)
    if language is None:
        out = avail_langs[name]
    elif language in ['all', 'all_nomatlab']:
        out = tuple(
            [x for x in avail_langs[name]
             if not (x.startswith('all')
                     or ((language == 'all_nomatlab') and (x == 'matlab')))]
            + list(base_langs[name]))
    else:
        out = tuple([language] + list(base_langs[name]))
    return out


def display_example(name, language, number_lines=False):
    r"""Display the yaml and source code for an example with syntax
    highlighting.

    Args:
        name (str): Name of the example.
        language (str): Language that example should be displayed in.
        number_lines (bool, optional): If True, line numbers will be added
            to the displayed examples. Defaults to False.

    """
    ex_yml = get_example_yaml(name, language)
    ex_src = get_example_source(name, language)
    tools.display_source(ex_yml, number_lines=number_lines)
    tools.display_source(ex_src, number_lines=number_lines)
    

__all__ = ['yamls', 'source']



import pprint
# Import classes for input/output channels
from yggdrasil.interface.YggInterface import (
    YggObjInput, YggObjOutput)

# Initialize input/output channels
in_channel = YggObjInput('inputB')
out_channel = YggObjOutput('outputB')

# Loop until there is no longer input or the queues are closed
while True:

    # Receive input from input channel
    # If there is an error, the flag will be False
    flag, obj = in_channel.recv()
    if not flag:
        print("Model B: No more input.")
        break

    # Print received message
    print('Model B: (%d verts, %d faces)' % (obj.nvert, obj.nface))
    pprint.pprint(obj)

    # Send output to output channel
    # If there is an error, the flag will be False
    flag = out_channel.send(obj)
    if not flag:
        raise RuntimeError("Model B: Error sending output.")
